@@ -42,6 +42,7 @@ reform_constrained_pars <- function(p, M, d, params, constraints=NULL, structura
   if(is.null(constraints)) { # For SGMVAR model with constrained structural parameters but no AR constraints
     q <- M*p*d^2
     psi_expanded <- params[(d*M + 1):(d*M + d^2*p*M)] # AR coefficients (without constraints)
+    psiNA <- FALSE
   } else {
     q <- ncol(constraints)
     psi <- params[(M*d + 1):(M*d + q)]
@@ -239,7 +240,7 @@ sort_components <- function(p, M, d, params, structural_pars=NULL) {
     if(all(ord == 1:(M - 1))) {
       return(params)
     } else {
-      n_zeros <- sum(structural_pars$W == 0)
+      n_zeros <- sum(structural_pars$W == 0, na.rm=TRUE)
       phi0 <- pick_phi0(p=p, M=M, d=d, params=params, structural_pars=structural_pars)
       A <- matrix(params[(d*M + 1):(d*M + d^2*p*M)], nrow=d^2*p, byrow=FALSE)
       lambdas <- matrix(params[(d*M + d^2*p*M + d^2 - n_zeros + 1):(d*M + d^2*p*M + d^2 - n_zeros + d*(M - 1))],
